@@ -24,14 +24,18 @@
 #include "funcapi.h"
 #include "miscadmin.h"
 
+#include "access/genam.h"
 #include "access/heapam.h"
 #include "access/htup_details.h"
+#include "access/parallel.h"
 
 #include "catalog/catalog.h"
 #include "catalog/indexing.h"
 #include "catalog/pg_auth_members.h"
 #include "catalog/pg_authid.h"
+#include "catalog/pg_db_role_setting.h"
 #include "commands/user.h"
+#include "common/int.h"
 #if PG_VERSION_NUM >= 140000
 #include "common/hmac.h"
 #endif
@@ -48,12 +52,16 @@
 #include "storage/shmem.h"
 #include "utils/acl.h"
 #include "utils/builtins.h"
+#include "utils/datetime.h"
+#include "utils/fmgroids.h"
 #include "utils/guc.h"
 #include "utils/rel.h"
 #include "utils/syscache.h"
 #include "utils/timestamp.h"
 #include "utils/varlena.h"
 
+#define NOT_IN_PARALLEL_WORKER (ParallelWorkerNumber < 0)
+
 /* Default passord encryption */
 #define Password_encryption = PASSWORD_TYPE_SCRAM_SHA_256;
 /* Name of external file to store password history in the PGDATA */
@@ -114,6 +122,7 @@ PG_MODULE_MAGIC;
 /* Hooks */
 static check_password_hook_type prev_check_password_hook = NULL;
 static ProcessUtility_hook_type prev_ProcessUtility = NULL;
+static ExecutorStart_hook_type prev_ExecutorStart = NULL;
 static shmem_startup_hook_type prev_shmem_startup_hook = NULL;
 #if PG_VERSION_NUM >= 150000
 static shmem_request_hook_type prev_shmem_request_hook = NULL;
@@ -182,6 +191,7 @@ static HTAB *pgaf_hash = NULL;
 extern void _PG_init(void);
 extern void _PG_fini(void);
 static void cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO);
+static void cc_ExecutorStart(QueryDesc *queryDesc, int eflags);
 
 static void flush_password_history(void);
 static pgphEntry *pgph_entry_alloc(pgphHashKey *key, TimestampTz password_date);
@@ -204,6 +214,8 @@ static float get_auth_failure(const char *username, Oid userid, int status);
 static float save_auth_failure(Port *port, Oid userid);
 static void remove_auth_failure(const char *username, Oid userid);
 static void pg_banned_role_internal(FunctionCallInfo fcinfo);
+static void set_force_change_password(Oid databaseid, Oid roleid, char *valuestr);
+static void reset_force_change_password(Oid databaseid, Oid roleid);
 
 /* Username flags*/
 static int username_min_length = 1;
@@ -231,8 +243,11 @@ static char *password_contain = NULL;
 static bool password_contain_username = true;
 static bool password_ignore_case = false;
 static int password_valid_until = 0;
+static int password_valid_warning = 0;
 static int password_valid_max = 0;
 static int auth_delay_milliseconds = 0;
+static bool password_change_first_login = false;
+static bool force_change_password = false;
 
 #if PG_VERSION_NUM >= 120000
 /*
@@ -264,15 +279,20 @@ static char *to_nlower(const char *str, size_t max) {
 }
 
 static bool str_contains(const char *chars, const char *str) {
-  for (const char *i = str; *i; i++) {
-    for (const char *j = chars; *j; j++) {
-      if (*i == *j) {
-        return true;
-      }
-    }
-  }
-
-  return false;
+    char *chars_copy;
+    char *token;
+    /* Make a copy of chars since strtok modifies the string */
+    chars_copy = pstrdup(chars);
+    
+    token = strtok(chars_copy, ",");
+    while (token != NULL) {
+        if (strstr(str, token) != NULL) {
+            return true;
+        }
+        token = strtok(NULL, ",");
+    }    
+    pfree(chars_copy);
+    return false;
 }
 
 static void check_str_counters(const char *str, int *lower, int *upper,
@@ -823,7 +843,9 @@ password_guc()
 
 	DefineCustomIntVariable("credcheck.password_valid_until",
 				gettext_noop("force use of VALID UNTIL clause in CREATE ROLE statement"
-					" with a minimum number of days"),
+					" with a minimum number of days or set it automatically to "
+					" now() + password_valid_until days in the ALTER ROLE statement"
+					" when the password is changed"),
 				NULL, &password_valid_until, 0, 0, INT_MAX,
 				PGC_SUSET, 0, NULL, NULL, NULL);
 
@@ -831,7 +853,23 @@ password_guc()
 				gettext_noop("force use of VALID UNTIL clause in CREATE ROLE statement"
 					" with a maximum number of days"),
 				NULL, &password_valid_max, 0, 0, INT_MAX,
+				PGC_SUSET, 1, NULL, NULL, NULL);
+
+	DefineCustomIntVariable("credcheck.password_valid_warning",
+				gettext_noop("throw a warning N days before the password expires"),
+				NULL, &password_valid_warning, 0, 0, INT_MAX,
 				PGC_SUSET, 0, NULL, NULL, NULL);
+
+	DefineCustomBoolVariable("credcheck.password_change_first_login",
+				gettext_noop("force the user to change his password at first login"),
+				NULL, &password_change_first_login, false, PGC_SUSET, 0,
+				NULL, NULL, NULL);
+
+	DefineCustomBoolVariable("credcheck_internal.force_change_password",
+				gettext_noop("force the user to change his password at first login"),
+				NULL, &force_change_password, false, PGC_SUSET, 0,
+				NULL, NULL, NULL);
+
 }
 
 #if PG_VERSION_NUM >= 120000
@@ -1157,7 +1195,7 @@ check_password_reuse(const char *username, const char *password)
 				/* mark that the password hash was found in the history */
 				found = true;
 
-				/* Check the password age again the reuse interval */
+				/* Check the password age against the reuse interval */
 				if (password_reuse_interval > 0)
 				{
 					TimestampTz       dt_now = GetCurrentTimestamp();
@@ -1344,6 +1382,7 @@ _PG_init(void)
 
 #if PG_VERSION_NUM < 150000
 	EmitWarningsOnPlaceholders("credcheck");
+	EmitWarningsOnPlaceholders("credcheck_internal");
 
         /*
          * Request additional shared resources.  (These are no-ops if we're not in
@@ -1356,6 +1395,7 @@ _PG_init(void)
         RequestNamedLWLockTranche(PGAF_TRANCHE_NAME, 1);
 #else
 	MarkGUCPrefixReserved("credcheck");
+	MarkGUCPrefixReserved("credcheck_internal");
 
 #endif
 
@@ -1376,6 +1416,9 @@ _PG_init(void)
 
 	prev_ClientAuthentication = ClientAuthentication_hook;
 	ClientAuthentication_hook = credcheck_max_auth_failure;
+
+	prev_ExecutorStart = ExecutorStart_hook;
+	ExecutorStart_hook = cc_ExecutorStart;
 }
 
 void
@@ -1395,194 +1438,298 @@ _PG_fini(void)
 static void
 cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 {
-	Node *parsetree = pstmt->utilityStmt;
+	if (NOT_IN_PARALLEL_WORKER)
+	{
+		Node *parsetree = pstmt->utilityStmt;
 
-	/* Execute the utility command before */
+		/* at first login we don't allow anything else than password change */
+		/*
+		 * When first login we don't allow anything else than password change.
+		 * Command \password issue a "show password_encryption" query after
+		 * change password prompts, allow it. The \password command will be
+		 * rejected anyway because it uses encrypted password.
+		 */
+		if (!is_in_whitelist(MyProcPort->user_name, username_whitelist))
+		{
+			if (nodeTag(parsetree) != T_AlterRoleStmt && force_change_password
+				&& strcmp(debug_query_string, "show password_encryption") != 0)
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+							errmsg(gettext_noop("you must change your password first."))));
+		}
+		statement_has_password = false;
+
+		switch (nodeTag(parsetree))
+		{
+			/* Intercept ALTER USER .. RENAME statements */
+			case T_RenameStmt:
+			{
+				RenameStmt *stmt = (RenameStmt *)parsetree;
+
+				/* We only take care of user renaming */
+				if (stmt->renameType == OBJECT_ROLE && stmt->newname != NULL)
+				{
+					if (is_in_whitelist(stmt->newname, username_whitelist) || is_in_whitelist(stmt->subname, username_whitelist))
+						break;
+				
+					/* check the validity of the username */
+					username_check(stmt->newname, NULL);
+
+#if PG_VERSION_NUM >= 120000
+					/* rename the user in the history table */
+					rename_user_in_history(stmt->subname, stmt->newname);
+#endif
+				}
+				break;
+			}
+
+			case T_AlterRoleStmt:
+			{
+				AlterRoleStmt *stmt = (AlterRoleStmt *)parsetree;
+				ListCell      *option;
+				char          *password;
+				bool           save_password = false;
+				DefElem    *dvalidUntil = NULL;
+				DefElem    *dpassword = NULL;
+
+				if (is_in_whitelist(stmt->role->rolename, username_whitelist))
+					break;
+
+				/* Extract options from the statement node tree */
+				foreach(option, stmt->options)
+				{
+					DefElem    *defel = (DefElem *) lfirst(option);
+
+					if (strcmp(defel->defname, "password") == 0)
+					{
+						dpassword = defel;
+					}
+					else if (strcmp(defel->defname, "validUntil") == 0)
+					{
+						dvalidUntil = defel;
+					}
+				}
+
+				if (dpassword == NULL && force_change_password)
+						ereport(ERROR,
+							(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+								errmsg(gettext_noop("you must change your password first."))));
+
+#if PG_VERSION_NUM >= 120000
+				/* check the password set */
+				if (dpassword && dpassword->arg)
+				{
+					statement_has_password = true;
+					password = strVal(dpassword->arg);
+					save_password = check_password_reuse(stmt->role->rolename, password);
+				}
+#endif
+				/*
+				 * when the user change his password, automatically set the valid until
+				 * date to now() + password_valid_until days if password_valid_until is set.
+				 */
+				if (!dvalidUntil && password_valid_until > 0)
+				{
+					Timestamp dt_now = GetCurrentTimestamp();
+					struct pg_tm tt, *tm = &tt;
+					fsec_t          fsec;
+					int             julian;
+					char           *validuntil;
+
+					if (timestamp2tm(dt_now, NULL, tm, &fsec, NULL, NULL) != 0)
+						ereport(ERROR,
+								(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+								 errmsg("timestamp out of range")));
+
+					/*
+					 * Add credcheck.password_valid_until days by converting to and from Julian.
+					 */
+					julian = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday);
+					if (pg_add_s32_overflow(julian, password_valid_until+1, &julian) ||
+						julian < 0)
+						ereport(ERROR,
+								(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+								 errmsg("timestamp out of range")));
+					j2date(julian, &tm->tm_year, &tm->tm_mon, &tm->tm_mday);
+					validuntil = malloc(sizeof(char)*11);
+					sprintf(validuntil, "%d-%d-%d", tm->tm_year, tm->tm_mon, tm->tm_mday);
+					dvalidUntil = makeDefElem("validUntil", (Node *) makeString(validuntil), -1);
+					((AlterRoleStmt *)parsetree)->options = lappend(((AlterRoleStmt *)parsetree)->options, dvalidUntil);
+				}
+
+				/* when a valid until date is set check that it is > to password_valid_until */
+				if (dvalidUntil && dvalidUntil->arg && password_valid_until > 0)
+				{
+					int valid_until = check_valid_until(strVal(dvalidUntil->arg));
+					if (valid_until < password_valid_until)
+						ereport(ERROR,
+							(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+								errmsg(gettext_noop("the VALID UNTIL option must have a date older than %d days"), password_valid_until)));
+				}
+				/* check that the valid until date is not under the limit of days */
+				if (dvalidUntil && dvalidUntil->arg && password_valid_max > 0)
+				{
+					int valid_max = check_valid_until(strVal(dvalidUntil->arg));
+					if (valid_max > password_valid_max)
+						ereport(ERROR,
+							(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+								errmsg(gettext_noop("the VALID UNTIL option must NOT have a date beyond %d days"), password_valid_max)));
+				}
+
+#if PG_VERSION_NUM >= 120000
+				/* The password can be saved into the history */
+				if (save_password)
+					save_password_in_history(stmt->role->rolename, password);
+#endif
+				if (force_change_password)
+				{
+					reset_force_change_password(0, get_role_oid(stmt->role->rolename, true));
+					force_change_password = false;
+				}
+
+				break;
+			}
+
+			case T_CreateRoleStmt:
+			{
+				CreateRoleStmt *stmt = (CreateRoleStmt *)parsetree;
+				ListCell       *option;
+				int             valid_until = 0;
+				int             valid_max = 0;
+				bool            has_valid_until = false; 
+				bool            save_password = false;
+				char           *password;
+				DefElem    *dpassword = NULL;
+				DefElem    *dvalidUntil = NULL;
+
+				if (is_in_whitelist(stmt->role, username_whitelist))
+					break;
+
+				/* check the validity of the username */
+				username_check(stmt->role, NULL);
+
+				/* Extract options from the statement node tree */
+				foreach(option, stmt->options)
+				{
+					DefElem    *defel = (DefElem *) lfirst(option);
+
+					if (strcmp(defel->defname, "password") == 0)
+					{
+						dpassword = defel;
+					}
+					else if (strcmp(defel->defname, "validUntil") == 0)
+					{
+						dvalidUntil = defel;
+					}
+				}
+#if PG_VERSION_NUM >= 120000
+				if (dpassword && dpassword->arg)
+				{
+					statement_has_password = true;
+					password = strVal(dpassword->arg);
+					save_password = check_password_reuse(stmt->role, password);
+				}
+#endif
+				/*
+				 * At user creation automatically set the valid until date to now() + password_valid_until
+				 * days if password_valid_until is set.
+				 */
+				if (!dvalidUntil && password_valid_until > 0)
+				{
+					Timestamp dt_now = GetCurrentTimestamp();
+					struct pg_tm tt, *tm = &tt;
+					fsec_t          fsec;
+					int             julian;
+					char            *validuntil;
+
+					if (timestamp2tm(dt_now, NULL, tm, &fsec, NULL, NULL) != 0)
+						ereport(ERROR,
+								(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+								 errmsg("timestamp out of range")));
+
+					/*
+					 * Add credcheck.password_valid_until days by converting to and from Julian.
+					 */
+					julian = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday);
+					if (pg_add_s32_overflow(julian, password_valid_until+1, &julian) ||
+						julian < 0)
+						ereport(ERROR,
+								(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+								 errmsg("timestamp out of range")));
+					j2date(julian, &tm->tm_year, &tm->tm_mon, &tm->tm_mday);
+					validuntil = malloc(sizeof(char)*11);
+					sprintf(validuntil, "%d-%d-%d", tm->tm_year, tm->tm_mon, tm->tm_mday);
+					dvalidUntil = makeDefElem("validUntil", (Node *) makeString(validuntil), 1);
+					((CreateRoleStmt *)parsetree)->options = lappend(((CreateRoleStmt *)parsetree)->options, dvalidUntil);
+				}
+
+				if (dvalidUntil && dvalidUntil->arg && password_valid_until > 0)
+				{
+					valid_until = check_valid_until(strVal(dvalidUntil->arg));
+					has_valid_until = true;
+				}
+				if (dvalidUntil && dvalidUntil->arg && password_valid_max > 0)
+				{
+					valid_max = check_valid_until(strVal(dvalidUntil->arg));
+					has_valid_until = true;
+				}
+
+				/* check that a VALID UNTIL option is present */
+				if ( !has_valid_until && (password_valid_until > 0 || password_valid_max > 0) )
+					ereport(ERROR,
+						(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+							errmsg(gettext_noop("require a VALID UNTIL option"))));
+
+				/* check that a minimum number of days for password validity is defined */
+				if (password_valid_until > 0 && valid_until < password_valid_until)
+					ereport(ERROR,
+						(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+							errmsg(gettext_noop("require a VALID UNTIL option with a date older than %d days"), password_valid_until)));
+
+				/* check that a maximum number of days for password validity is defined */
+				if (password_valid_max > 0 && valid_max < password_valid_max)
+					ereport(ERROR,
+						(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+							errmsg(gettext_noop("require a VALID UNTIL option with a date beyond %d days"), password_valid_max)));
+
+				/* set force change password option if password_change_first_login is set */
+				if (password_change_first_login)
+					set_force_change_password(0, get_role_oid(stmt->role, true), "true");
+
+#if PG_VERSION_NUM >= 120000
+				/* The password can be saved into the history */
+				if (save_password)
+					save_password_in_history(stmt->role, password);
+#endif
+				break;
+			}
+
+#if PG_VERSION_NUM >= 120000
+			case T_DropRoleStmt:
+			{
+				DropRoleStmt *stmt = (DropRoleStmt *)parsetree;
+				ListCell   *item;
+
+				foreach(item, stmt->roles)
+				{
+					RoleSpec   *rolspec = lfirst(item);
+
+					remove_user_from_history(rolspec->rolename);
+				}
+				break;
+			}
+#endif
+			default:
+				break;
+		}
+	}
+
+	/* Execute the utility command now */
 	if (prev_ProcessUtility)
 		prev_ProcessUtility(PEL_PROCESSUTILITY_ARGS);
 	else
 		standard_ProcessUtility(PEL_PROCESSUTILITY_ARGS);
 
-	statement_has_password = false;
-
-	switch (nodeTag(parsetree))
-	{
-		/* Intercept ALTER USER .. RENAME statements */
-		case T_RenameStmt:
-		{
-			RenameStmt *stmt = (RenameStmt *)parsetree;
-			/* We only take care of user renaming */
-			if (stmt->renameType == OBJECT_ROLE && stmt->newname != NULL)
-			{
-				if (is_in_whitelist(stmt->newname, username_whitelist) || is_in_whitelist(stmt->subname, username_whitelist))
-					break;
-			
-				/* check the validity of the username */
-				username_check(stmt->newname, NULL);
-
-#if PG_VERSION_NUM >= 120000
-				/* rename the user in the history table */
-				rename_user_in_history(stmt->subname, stmt->newname);
-#endif
-			}
-			break;
-		}
-
-		case T_AlterRoleStmt:
-		{
-			AlterRoleStmt *stmt = (AlterRoleStmt *)parsetree;
-			ListCell      *option;
-			char          *password;
-			bool           save_password = false;
-			DefElem    *dvalidUntil = NULL;
-			DefElem    *dpassword = NULL;
-
-			if (is_in_whitelist(stmt->role->rolename, username_whitelist))
-				break;
-
-			/* Extract options from the statement node tree */
-			foreach(option, stmt->options)
-			{
-				DefElem    *defel = (DefElem *) lfirst(option);
-
-				if (strcmp(defel->defname, "password") == 0)
-				{
-					dpassword = defel;
-				}
-				else if (strcmp(defel->defname, "validUntil") == 0)
-				{
-					dvalidUntil = defel;
-				}
-			}
-
-#if PG_VERSION_NUM >= 120000
-			if (dpassword && dpassword->arg)
-			{
-				statement_has_password = true;
-				password = strVal(dpassword->arg);
-				save_password = check_password_reuse(stmt->role->rolename, password);
-			}
-#endif
-			if (dvalidUntil && dvalidUntil->arg && password_valid_until > 0)
-			{
-				int valid_until = check_valid_until(strVal(dvalidUntil->arg));
-				if (valid_until < password_valid_until)
-					ereport(ERROR,
-						(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
-							errmsg(gettext_noop("the VALID UNTIL option must have a date older than %d days"), password_valid_until)));
-			}
-			if (dvalidUntil && dvalidUntil->arg && password_valid_max > 0)
-			{
-				int valid_max = check_valid_until(strVal(dvalidUntil->arg));
-				if (valid_max > password_valid_max)
-					ereport(ERROR,
-						(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
-							errmsg(gettext_noop("the VALID UNTIL option must NOT have a date beyond %d days"), password_valid_max)));
-			}
-
-#if PG_VERSION_NUM >= 120000
-			/* The password can be saved into the history */
-			if (save_password)
-				save_password_in_history(stmt->role->rolename, password);
-#endif
-			break;
-		}
-
-		case T_CreateRoleStmt:
-		{
-			CreateRoleStmt *stmt = (CreateRoleStmt *)parsetree;
-			ListCell       *option;
-			int             valid_until = 0;
-			int             valid_max = 0;
-			bool            has_valid_until = false; 
-			bool            save_password = false;
-			char           *password;
-			DefElem    *dpassword = NULL;
-			DefElem    *dvalidUntil = NULL;
-
-			if (is_in_whitelist(stmt->role, username_whitelist))
-				break;
-
-			/* check the validity of the username */
-			username_check(stmt->role, NULL);
-
-			/* Extract options from the statement node tree */
-			foreach(option, stmt->options)
-			{
-				DefElem    *defel = (DefElem *) lfirst(option);
-
-				if (strcmp(defel->defname, "password") == 0)
-				{
-					dpassword = defel;
-				}
-				else if (strcmp(defel->defname, "validUntil") == 0)
-				{
-					dvalidUntil = defel;
-				}
-			}
-
-#if PG_VERSION_NUM >= 120000
-			if (dpassword && dpassword->arg)
-			{
-				statement_has_password = true;
-				password = strVal(dpassword->arg);
-				save_password = check_password_reuse(stmt->role, password);
-			}
-#endif
-			if (dvalidUntil && dvalidUntil->arg && password_valid_until > 0)
-			{
-				valid_until = check_valid_until(strVal(dvalidUntil->arg));
-				has_valid_until = true;
-			}
-			if (dvalidUntil && dvalidUntil->arg && password_valid_max > 0)
-			{
-				valid_max = check_valid_until(strVal(dvalidUntil->arg));
-				has_valid_until = true;
-			}
-
-			/* check that a VALID UNTIL option is present */
-			if ( !has_valid_until && (password_valid_until > 0 || password_valid_max > 0) )
-				ereport(ERROR,
-					(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
-						errmsg(gettext_noop("require a VALID UNTIL option"))));
-
-			/* check that a minimum number of days for password validity is defined */
-			if (password_valid_until > 0 && valid_until < password_valid_until)
-				ereport(ERROR,
-					(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
-						errmsg(gettext_noop("require a VALID UNTIL option with a date older than %d days"), password_valid_until)));
-
-			/* check that a maximum number of days for password validity is defined */
-			if (password_valid_max > 0 && valid_max > password_valid_max)
-				ereport(ERROR,
-					(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
-						errmsg(gettext_noop("require a VALID UNTIL option with a date beyond %d days"), password_valid_max)));
-			/* The password can be saved into the history */
-#if PG_VERSION_NUM >= 120000
-			if (save_password)
-				save_password_in_history(stmt->role, password);
-#endif
-			break;
-		}
-
-#if PG_VERSION_NUM >= 120000
-		case T_DropRoleStmt:
-		{
-			DropRoleStmt *stmt = (DropRoleStmt *)parsetree;
-			ListCell   *item;
-
-			foreach(item, stmt->roles)
-			{
-				RoleSpec   *rolspec = lfirst(item);
-
-				remove_user_from_history(rolspec->rolename);
-			}
-			break;
-		}
-#endif
-		default:
-			break;
-	}
 }
 
 #if PG_VERSION_NUM >= 120000
@@ -2506,5 +2653,167 @@ pg_banned_role_internal(FunctionCallInfo fcinfo)
 	LWLockRelease(pgaf->lock);
 }
 
+static void
+cc_ExecutorStart(QueryDesc *queryDesc, int eflags)
+{
+        elog(DEBUG1, "cc_ExecutorStart()");
 
+	if (NOT_IN_PARALLEL_WORKER)
+	{
+		/*
+		 * When first login we don't allow anything else than password change.
+		 * This rule is also checked in the ProcessUtility hook because we need
+		 * to allow the ALTER ROLE command to change the password. We must allow
+		 * SELECT CURRENT_USER which is sent by the \passwd command before
+		 * password change.
+		 */
+		if (debug_query_string != NULL && strcmp(debug_query_string, "SELECT CURRENT_USER") != 0)
+		{
+			if (!is_in_whitelist(MyProcPort->user_name, username_whitelist))
+			{
+				if (queryDesc->operation != CMD_UTILITY && force_change_password)
+						ereport(ERROR,
+							(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+								errmsg(gettext_noop("you must change your password first."))));
+			}
+		}
+	}
+
+        /* Continue the normal behavior */
+        if (prev_ExecutorStart)
+                prev_ExecutorStart(queryDesc, eflags);
+        else
+                standard_ExecutorStart(queryDesc, eflags);
+
+	elog(DEBUG1, "End of cc_ExecutorStart()");
+}
+
+static void
+set_force_change_password(Oid databaseid, Oid roleid, char *valuestr)
+{
+	//HeapTuple       tuple;
+	Relation	rel;
+	ScanKeyData scankey[2];
+	SysScanDesc scan;
+
+	/* Get the old tuple, if any. */
+
+	rel = table_open(DbRoleSettingRelationId, RowExclusiveLock);
+	ScanKeyInit(&scankey[0],
+				Anum_pg_db_role_setting_setdatabase,
+				BTEqualStrategyNumber, F_OIDEQ,
+				ObjectIdGetDatum(databaseid));
+	ScanKeyInit(&scankey[1],
+				Anum_pg_db_role_setting_setrole,
+				BTEqualStrategyNumber, F_OIDEQ,
+				ObjectIdGetDatum(roleid));
+	scan = systable_beginscan(rel, DbRoleSettingDatidRolidIndexId, true,
+							  NULL, 2, scankey);
+	//tuple = systable_getnext(scan);
+
+	if (valuestr)
+	{
+		/* non-null valuestr means it's not RESET, so insert a new tuple */
+		HeapTuple	newtuple;
+		Datum		values[Natts_pg_db_role_setting];
+		bool		nulls[Natts_pg_db_role_setting];
+		ArrayType  *a;
+
+		memset(nulls, false, sizeof(nulls));
+
+		a = GUCArrayAdd(NULL, "credcheck_internal.force_change_password", valuestr);
+
+		values[Anum_pg_db_role_setting_setdatabase - 1] = ObjectIdGetDatum(databaseid);
+		values[Anum_pg_db_role_setting_setrole - 1] = ObjectIdGetDatum(roleid);
+		values[Anum_pg_db_role_setting_setconfig - 1] = PointerGetDatum(a);
+		newtuple = heap_form_tuple(RelationGetDescr(rel), values, nulls);
+
+		CatalogTupleInsert(rel, newtuple);
+	}
+	systable_endscan(scan);
+
+	/* Close pg_db_role_setting, but keep lock till commit */
+	table_close(rel, NoLock);
+}
+
+static void
+reset_force_change_password(Oid databaseid, Oid roleid)
+{
+	bool need_priv_escalation = !superuser(); /* we might be a SU */
+	Oid     save_userid;
+	int     save_sec_context;
+	HeapTuple	tuple;
+	Relation	rel;
+	ScanKeyData scankey[2];
+	SysScanDesc scan;
+
+	/* The setting reseet must be done as SU */
+	if (need_priv_escalation)
+	{
+		/* Get current user's Oid and security context */
+		GetUserIdAndSecContext(&save_userid, &save_sec_context);
+		/* Become superuser */
+		SetUserIdAndSecContext(BOOTSTRAP_SUPERUSERID, save_sec_context
+							| SECURITY_LOCAL_USERID_CHANGE
+							| SECURITY_RESTRICTED_OPERATION);
+	}
+
+	/* Get the old tuple, if any. */
+	rel = table_open(DbRoleSettingRelationId, RowExclusiveLock);
+	ScanKeyInit(&scankey[0],
+				Anum_pg_db_role_setting_setdatabase,
+				BTEqualStrategyNumber, F_OIDEQ,
+				ObjectIdGetDatum(databaseid));
+	ScanKeyInit(&scankey[1],
+				Anum_pg_db_role_setting_setrole,
+				BTEqualStrategyNumber, F_OIDEQ,
+				ObjectIdGetDatum(roleid));
+	scan = systable_beginscan(rel, DbRoleSettingDatidRolidIndexId, true,
+							  NULL, 2, scankey);
+	tuple = systable_getnext(scan);
+
+	if (HeapTupleIsValid(tuple))
+	{
+		Datum		repl_val[Natts_pg_db_role_setting];
+		bool		repl_null[Natts_pg_db_role_setting];
+		bool		repl_repl[Natts_pg_db_role_setting];
+		HeapTuple	newtuple;
+		Datum		datum;
+		bool		isnull;
+		ArrayType  *a;
+
+		memset(repl_repl, false, sizeof(repl_repl));
+		repl_repl[Anum_pg_db_role_setting_setconfig - 1] = true;
+		repl_null[Anum_pg_db_role_setting_setconfig - 1] = false;
+
+		/* Extract old value of setconfig */
+		datum = heap_getattr(tuple, Anum_pg_db_role_setting_setconfig,
+							 RelationGetDescr(rel), &isnull);
+		a = isnull ? NULL : DatumGetArrayTypeP(datum);
+
+		a = GUCArrayDelete(a, "credcheck_internal.force_change_password");
+
+		if (a)
+		{
+			repl_val[Anum_pg_db_role_setting_setconfig - 1] =
+				PointerGetDatum(a);
+
+			newtuple = heap_modify_tuple(tuple, RelationGetDescr(rel),
+										 repl_val, repl_null, repl_repl);
+			CatalogTupleUpdate(rel, &tuple->t_self, newtuple);
+		}
+		else
+			CatalogTupleDelete(rel, &tuple->t_self);
+	}
+
+	systable_endscan(scan);
+
+	/* Close pg_db_role_setting, but keep lock till commit */
+	table_close(rel, NoLock);
+
+	/* Restore user's privileges */
+	if (need_priv_escalation)
+		SetUserIdAndSecContext(save_userid, save_sec_context);
+
+}
 
